/* Gibbs Sampling procedure using SMILE library 
   Author: just a girl
   1. Sampling 
   2. Inference  
 */

#include <iostream>
#include <cstdio>
#include <random>
#include <smile.h>
#include "smile_license.h"
using namespace std;

// helpers - generate sample point conditional on other nodes
static void changeEvidence(DSL_network &net, const char *nodeID, int outcomeState){
    DSL_node* node = net.GetNode(nodeID);
    DSL_idArray outcomes = node -> Def() -> GetOutcomeIds();
    
    const char* outcomeID = outcomes[outcomeState];
    node -> Val() -> SetEvidence(outcomeID);
}

void setParentEvidence(DSL_network &net, int handle, int outcomeState){
    const DSL_intArray &parents = net.GetParents(handle);
    

    for( int p: parents){
        const char *nodeID = net.GetNode(p) -> GetId();
        changeEvidence(net, *nodeID, outcomeState);
    }

}

int assignOutcome(DSL_network &net, int handle, double x){
    DSL_node* node = net.GetNode(handle);
    const char* nodeID = node -> GetId();
    const  DSL_Dmatrix& nodeProbabilities = *node -> Val() -> GetMatrix();
    int numStates = nodeProbabilities.GetSize();
    int testNumStates = node -> Def() -> GetNumberOfOutcomes();

    cout << "handle: " << handle << " Node: " << nodeID << endl;

    cout << "Assigning outcome for node " << nodeID << endl;
    if(numStates != testNumStates){
        cout << "Something is not right! NumStates = " << testNumStates << " while probability arraySize = " << numStates << " for " << nodeID << endl;

        for(int d = 0; d < nodeProbabilities.GetSize(); d++){
            const char *nodeID = node -> GetId();
            const DSL_idArray &outcomeIds = *node->Def()->GetOutcomeIds();
            cout << "P(" << nodeID << "=" << outcomeIds[d] << ") = " << nodeProbabilities[d] << endl;
        }

        exit(2);
    }

    
       
    double boundaries[numStates];
    int boundarySize = sizeof(boundaries)/sizeof(*boundaries);
    
    // establish boundary values for checking against simulated value x
    for(int b = 0; b < boundarySize; b++){
        double sum = 0;
        for(int v = 0; v <= b; v++){
            //cout <<  "Probability: " << probabilities[v] << endl;
            sum += nodeProbabilities[v];
        }
        boundaries[b] = sum;

    }

    /*for(int b = 0; b < boundarySize; b++){
        cout << "Boundary value: " << boundaries[b] << endl;
    }*/
    
    int assigned_outcome = 0;
    cout << "Num states: " << numStates << endl;
    while(assigned_outcome <= numStates){
        cout << "current assigned outcome: " << assigned_outcome << "\t simulated x = " << x << "\t Boundary value: " << boundaries[assigned_outcome] << endl;
        
        // if the simulated value is <= the boundary value, return the assigned outcome, otherwise increment and check again
        if(assigned_outcome == numStates){
            break;
        }else if(x <= boundaries[assigned_outcome]){
            break;
        }else{
            assigned_outcome++;
            //cout << "x not less than boundary "<< boundaries[assigned_outcome] << ", incrementing assigned outcome to " << assigned_outcome << endl;
        }

    }

    return(assigned_outcome);
}

int main(){

    // initialize/read in the network
    DSL_network theNetwork; 

    int res = theNetwork.ReadFile("sprinkler.xdsl");
    if(res != DSL_OKAY){
        cout << "ReadFile failed" <<endl;
        return res;
    }
    theNetwork.UpdateBeliefs();
    
    // initialize sampling parameters
    int n = 10; // number of samples
    int m = theNetwork.GetNumberOfNodes();
    int samples[n+1][m+1]; // nxm matrix to store samples
    double weights[n+1]; // calculated by probability
    std::default_random_engine generator;
    std::uniform_real_distribution<double> unif(0.0,1.0);

    //cout << "Number of nodes: " << m << endl;

    
    // make sure the nodes are in topological order
    DSL_intArray orderedNet = theNetwork.PartialOrdering();

    // get all nodes
    DSL_idArray theNodes;
    theNetwork.GetAllNodeIds(theNodes);

    // initialize first sample completely randomly
    for(int h = 0; h < orderedNet.GetSize(); h++){
        int handle = orderedNet[h];
        int num_states = theNetwork.GetNode(handle) -> Def() -> GetNumberOfOutcomes();
        
        //cout << "Node: " << theNetwork.GetNode(handle) -> GetId() << endl;
        samples[0][h] =  (rand() % num_states);
    }

    // generate other samples by randomly generating 1 node at a time
    for(int i = 1; i < n; i++){
        for (int h = 0; h < orderedNet.GetSize(); h++){
            //cout << "Node: " << theNetwork.GetNode(handle) -> GetId() << endl;
            //cout << "h = " << h << endl;

            // generate random sample for 1 node, holding others constant
            if(i > m){
                // an incredibly scuffed way to determine which node needs to be generated
                //cout << " i > m, (" << i << ">" << m << ") using modulus to identify node:" << endl;
                int noi = (i % m) -1;
                if(noi == -1){
                    noi = m - 1;
                }
                //cout << "Modulus i%m " << noi << endl;
                // This is the node that needs to be sampled
                if(noi == h){
                    samples[i][h] = 888; //test


    
                }else{
                    samples[i][h] = samples[i-1][h];
                }
            }else if(h == i - 1){
                
                //TODO:  reduce clutter by extracting code below to helper function generateSamplePoint
                int handle = orderedNet[h];
                DSL_node *node = theNetwork.GetNode(handle);
                const DSL_intArray &parents = node -> Def() ->GetParents();
                
                // Set evidence to parent values and update network before extracting probabilities
                if(parents.IsEmpty()){

                    // straight grab probabilitiy from CPT (no beliefs needed)
                    cout << "Node " << node -> GetId() << " has no parents, using probabilities directly from table" << endl;
                    samples[i][h] = assignOutcome(theNetwork, handle, unif(generator));
                    //samples[i][h] = h+10; //test

                // grab probabilities from the table if the node has no parents    
                }else{
                    // grab conditional probability from CPT based on values of parents in sample i
                    // or? set evidence and update beliefs, then extract probability? That will probably be easier tbh
                    //TODO: helper function setParentEvidence(network, node, parents)


                    //samples[i][h] = assignOutcome(node, unif(generator));
                    samples[i][h] = h+10; //test
                }
                
            

                /*int test_assigned_outcome = assignOutcome(node, unif(generator));
                cout << "test assigned outcome: " << test_assigned_outcome << endl;
                samples[i][h] = test_assigned_outcome;*/


            }else{
                samples[i][h] = samples[i-1][h];
            }

            // reset the board after every node
            theNetwork.ClearAllEvidence();
            theNetwork.UpdateBeliefs();
        }

    }


// print samples 
for(int i = 0; i < theNodes.GetSize(); i++){
    cout << theNodes[i] << "\t";
}
cout << endl;
for(int i = 0; i < n; i++){
    for(int j = 0; j < m; j++){
        cout << samples[i][j] << "\t";
    }
    cout << endl;
}

    return 0;
}